import tensorflow as tf
import six


def get_shape_list(tensor, expected_rank=None, name=None):
    """Returns a list of the shape of tensor, preferring static dimensions.

    Args:
      tensor: A tf.Tensor object to find the shape of.
      expected_rank: (optional) int. The expected rank of `tensor`. If this is
        specified and the `tensor` has a different rank, and exception will be
        thrown.
      name: Optional name of the tensor for the error message.

    Returns:
      A list of dimensions of the shape of tensor. All static dimensions will
      be returned as python integers, and dynamic dimensions will be returned
      as tf.Tensor scalars.
    """
    if expected_rank is not None:
        assert_rank(tensor, expected_rank, name)

    shape = tensor.shape.as_list()

    non_static_indexes = []
    for (index, dim) in enumerate(shape):
        if dim is None:
            non_static_indexes.append(index)

    if not non_static_indexes:
        return shape

    dyn_shape = tf.shape(tensor)
    for index in non_static_indexes:
        shape[index] = dyn_shape[index]
    return shape


def assert_rank(tensor, expected_rank, name=None):
    """Raises an exception if the tensor rank is not of the expected rank.

    Args:
      tensor: A tf.Tensor to check the rank of.
      expected_rank: Python integer or list of integers, expected rank.
      name: Optional name of the tensor for the error message.

    Raises:
      ValueError: If the expected shape doesn't match the actual shape.
    """
    expected_rank_dict = {}
    if isinstance(expected_rank, six.integer_types):
        expected_rank_dict[expected_rank] = True
    else:
        for x in expected_rank:
            expected_rank_dict[x] = True

    actual_rank = tensor.shape.ndims
    if actual_rank not in expected_rank_dict:
        raise ValueError(
            "For the tensor `%s`, the actual tensor rank `%d` (shape = %s) is not "
            "equal to the expected tensor rank `%s`" %
            (name, actual_rank, str(tensor.shape), str(expected_rank)))


@tf.keras.utils.register_keras_serializable(package='Text')
def gelu(x):
    """Gaussian Error Linear Unit.

    This is a smoother version of the RELU.
    Original paper: https://arxiv.org/abs/1606.08415
    Args:
      x: float Tensor to perform activation.

    Returns:
      `x` with the GELU activation applied.
    """
    return tf.keras.activations.gelu(x, approximate=True)



import hashlib
import os
import sys
import logging
import time
import csv

try:
    from pathlib import Path
except ImportError:
    Path = None

log_stderr = logging.StreamHandler(sys.stderr)
default_logger = logging.getLogger(__name__)
default_logger.setLevel(logging.DEBUG)
default_logger.addHandler(log_stderr)

package_path = os.path.abspath(os.path.dirname(sys.modules[__package__].__file__))


def sha1(query: str) -> str:
    m = hashlib.sha1()
    m.update(query.encode())
    return m.hexdigest()


def md5(query: str) -> str:
    m = hashlib.md5()
    m.update(query.encode())
    return m.hexdigest()


def read_lines(filename, func=None):
    if Path is not None and isinstance(filename, Path):
        filename = str(filename)
    if hasattr(filename, "read"):
        f_obj = filename
    else:
        f_obj = open(filename, 'r', encoding='utf-8')
    for l in f_obj:
        l = l.strip()
        if func is not None:
            l = func(l)
        yield l
    f_obj.close()


def read_csv_row(filename, func=None):
    if Path is not None and isinstance(filename, Path):
        filename = str(filename)
    if hasattr(filename, "read"):
        f_obj = filename
    else:
        f_obj = open(filename, 'r', encoding='utf-8')
    f_csv = csv.reader(f_obj)
    for row in f_csv:
        if func is not None:
            row = func(row)
        yield row
    f_obj.close()


def logging_run_time(func):
    def wrapper(*args, **kw):
        local_time = time.time()
        default_logger.info('Function [%s] start' % func.__name__)
        t = func(*args, **kw)
        default_logger.info('Function [%s] cost time %.2f second' % (func.__name__, time.time() - local_time))
        return t

    return wrapper


if __name__ == '__main__':

    with open("/home/geb/PycharmProjects/whisper-data-dependence/online/preprocess/convertZ.txt", 'r',
              encoding='utf-8') as f, open(
        "/home/geb/PycharmProjects/whisper-data-dependence/preprocess/convertSpecial.txt", 'r',
        encoding='utf-8') as f2, open("data/letter_mapping.txt", 'w', encoding='utf-8') as f3:
        for line in f:
            a = line.strip().split()
            f3.write(f"{md5(a[0])}\t{a[1]}\t1\n")

        for line in f2:
            a = line.strip().split()
            f3.write(f"{md5(a[0])}\t{a[1]}\t0\n")



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Keras-based TransformerEncoder block layer."""

import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="xdnlp")
class TransformerEncoderBlock(tf.keras.layers.Layer):
    """TransformerEncoderBlock layer.

    This layer implements the Transformer Encoder from
    "Attention Is All You Need". (https://arxiv.org/abs/1706.03762),
    which combines a `tf.keras.layers.MultiHeadAttention` layer with a
    two-layer feedforward network.

    References:
      [Attention Is All You Need](https://arxiv.org/abs/1706.03762)
      [BERT: Pre-training of Deep Bidirectional Transformers for Language
       Understanding](https://arxiv.org/abs/1810.04805)
    """

    def __init__(self,
                 num_attention_heads,
                 inner_dim,
                 inner_activation,
                 output_range=None,
                 kernel_initializer="glorot_uniform",
                 bias_initializer="zeros",
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 use_bias=True,
                 norm_first=False,
                 norm_epsilon=1e-12,
                 output_dropout=0.0,
                 attention_dropout=0.0,
                 inner_dropout=0.0,
                 attention_initializer=None,
                 attention_axes=None,
                 **kwargs):
        """Initializes `TransformerEncoderBlock`.

        Args:
          num_attention_heads: Number of attention heads.
          inner_dim: The output dimension of the first Dense layer in a two-layer
            feedforward network.
          inner_activation: The activation for the first Dense layer in a two-layer
            feedforward network.
          output_range: the sequence output range, [0, output_range) for slicing the
            target sequence. `None` means the target sequence is not sliced.
          kernel_initializer: Initializer for dense layer kernels.
          bias_initializer: Initializer for dense layer biases.
          kernel_regularizer: Regularizer for dense layer kernels.
          bias_regularizer: Regularizer for dense layer biases.
          activity_regularizer: Regularizer for dense layer activity.
          kernel_constraint: Constraint for dense layer kernels.
          bias_constraint: Constraint for dense layer kernels.
          use_bias: Whether to enable use_bias in attention layer. If set False,
            use_bias in attention layer is disabled.
          norm_first: Whether to normalize inputs to attention and intermediate
            dense layers. If set False, output of attention and intermediate dense
            layers is normalized.
          norm_epsilon: Epsilon value to initialize normalization layers.
          output_dropout: Dropout probability for the post-attention and output
            dropout.
          attention_dropout: Dropout probability for within the attention layer.
          inner_dropout: Dropout probability for the first Dense layer in a
            two-layer feedforward network.
          attention_initializer: Initializer for kernels of attention layers. If set
            `None`, attention layers use kernel_initializer as initializer for
            kernel.
          attention_axes: axes over which the attention is applied. `None` means
            attention over all axes, but batch, heads, and features.
          **kwargs: keyword arguments/
        """
        super().__init__(**kwargs)

        self._num_heads = num_attention_heads
        self._inner_dim = inner_dim
        self._inner_activation = inner_activation
        self._attention_dropout = attention_dropout
        self._attention_dropout_rate = attention_dropout
        self._output_dropout = output_dropout
        self._output_dropout_rate = output_dropout
        self._output_range = output_range
        self._kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self._bias_initializer = tf.keras.initializers.get(bias_initializer)
        self._kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)
        self._bias_regularizer = tf.keras.regularizers.get(bias_regularizer)
        self._activity_regularizer = tf.keras.regularizers.get(activity_regularizer)
        self._kernel_constraint = tf.keras.constraints.get(kernel_constraint)
        self._bias_constraint = tf.keras.constraints.get(bias_constraint)
        self._use_bias = use_bias
        self._norm_first = norm_first
        self._norm_epsilon = norm_epsilon
        self._inner_dropout = inner_dropout
        if attention_initializer:
            self._attention_initializer = tf.keras.initializers.get(
                attention_initializer)
        else:
            self._attention_initializer = self._kernel_initializer
        self._attention_axes = attention_axes

    def build(self, input_shape):
        if isinstance(input_shape, tf.TensorShape):
            input_tensor_shape = input_shape
        elif isinstance(input_shape, (list, tuple)):
            input_tensor_shape = tf.TensorShape(input_shape[0])
        else:
            raise ValueError(
                "The type of input shape argument is not supported, got: %s" %
                type(input_shape))
        einsum_equation = "abc,cd->abd"
        if len(input_tensor_shape.as_list()) > 3:
            einsum_equation = "...bc,cd->...bd"
        hidden_size = input_tensor_shape[-1]
        if hidden_size % self._num_heads != 0:
            raise ValueError(
                "The input size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (hidden_size, self._num_heads))
        self._attention_head_size = int(hidden_size // self._num_heads)
        common_kwargs = dict(
            bias_initializer=self._bias_initializer,
            kernel_regularizer=self._kernel_regularizer,
            bias_regularizer=self._bias_regularizer,
            activity_regularizer=self._activity_regularizer,
            kernel_constraint=self._kernel_constraint,
            bias_constraint=self._bias_constraint)
        self._attention_layer = tf.keras.layers.MultiHeadAttention(
            num_heads=self._num_heads,
            key_dim=self._attention_head_size,
            dropout=self._attention_dropout,
            use_bias=self._use_bias,
            kernel_initializer=self._attention_initializer,
            attention_axes=self._attention_axes,
            name="self_attention",
            **common_kwargs)
        self._attention_dropout = tf.keras.layers.Dropout(rate=self._output_dropout)
        # Use float32 in layernorm for numeric stability.
        # It is probably safe in mixed_float16, but we haven't validated this yet.
        self._attention_layer_norm = (
            tf.keras.layers.LayerNormalization(
                name="self_attention_layer_norm",
                axis=-1,
                epsilon=self._norm_epsilon,
                dtype=tf.float32))
        self._intermediate_dense = tf.keras.layers.experimental.EinsumDense(
            einsum_equation,
            output_shape=(None, self._inner_dim),
            bias_axes="d",
            kernel_initializer=self._kernel_initializer,
            name="intermediate",
            **common_kwargs)
        policy = tf.keras.mixed_precision.global_policy()
        if policy.name == "mixed_bfloat16":
            # bfloat16 causes BERT with the LAMB optimizer to not converge
            # as well, so we use float32.
            # TODO(b/154538392): Investigate this.
            policy = tf.float32
        self._intermediate_activation_layer = tf.keras.layers.Activation(
            self._inner_activation, dtype=policy)
        self._inner_dropout_layer = tf.keras.layers.Dropout(
            rate=self._inner_dropout)
        self._output_dense = tf.keras.layers.experimental.EinsumDense(
            einsum_equation,
            output_shape=(None, hidden_size),
            bias_axes="d",
            name="output",
            kernel_initializer=self._kernel_initializer,
            **common_kwargs)
        self._output_dropout = tf.keras.layers.Dropout(rate=self._output_dropout)
        # Use float32 in layernorm for numeric stability.
        self._output_layer_norm = tf.keras.layers.LayerNormalization(
            name="output_layer_norm",
            axis=-1,
            epsilon=self._norm_epsilon,
            dtype=tf.float32)

        super(TransformerEncoderBlock, self).build(input_shape)

    def get_config(self):
        config = {
            "num_attention_heads":
                self._num_heads,
            "inner_dim":
                self._inner_dim,
            "inner_activation":
                self._inner_activation,
            "output_dropout":
                self._output_dropout_rate,
            "attention_dropout":
                self._attention_dropout_rate,
            "output_range":
                self._output_range,
            "kernel_initializer":
                tf.keras.initializers.serialize(self._kernel_initializer),
            "bias_initializer":
                tf.keras.initializers.serialize(self._bias_initializer),
            "kernel_regularizer":
                tf.keras.regularizers.serialize(self._kernel_regularizer),
            "bias_regularizer":
                tf.keras.regularizers.serialize(self._bias_regularizer),
            "activity_regularizer":
                tf.keras.regularizers.serialize(self._activity_regularizer),
            "kernel_constraint":
                tf.keras.constraints.serialize(self._kernel_constraint),
            "bias_constraint":
                tf.keras.constraints.serialize(self._bias_constraint),
            "use_bias":
                self._use_bias,
            "norm_first":
                self._norm_first,
            "norm_epsilon":
                self._norm_epsilon,
            "inner_dropout":
                self._inner_dropout,
            "attention_initializer":
                tf.keras.initializers.serialize(self._attention_initializer),
            "attention_axes": self._attention_axes,
        }
        base_config = super(TransformerEncoderBlock, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def call(self, inputs):
        """Transformer self-attention encoder block call.

        Args:
          inputs: a single tensor or a list of tensors.
            `input tensor` as the single sequence of embeddings.
            [`input tensor`, `attention mask`] to have the additional attention
              mask.
            [`query tensor`, `key value tensor`, `attention mask`] to have separate
              input streams for the query, and key/value to the multi-head
              attention.

        Returns:
          An output tensor with the same dimensions as input/query tensor.
        """
        if isinstance(inputs, (list, tuple)):
            if len(inputs) == 2:
                input_tensor, attention_mask = inputs
                key_value = None
            elif len(inputs) == 3:
                input_tensor, key_value, attention_mask = inputs
            else:
                raise ValueError("Unexpected inputs to %s with length at %d" %
                                 (self.__class__, len(inputs)))
        else:
            input_tensor, key_value, attention_mask = (inputs, None, None)

        if self._output_range:
            if self._norm_first:
                source_tensor = input_tensor[:, 0:self._output_range, :]
                input_tensor = self._attention_layer_norm(input_tensor)
                if key_value is not None:
                    key_value = self._attention_layer_norm(key_value)
            target_tensor = input_tensor[:, 0:self._output_range, :]
            if attention_mask is not None:
                attention_mask = attention_mask[:, 0:self._output_range, :]
        else:
            if self._norm_first:
                source_tensor = input_tensor
                input_tensor = self._attention_layer_norm(input_tensor)
                if key_value is not None:
                    key_value = self._attention_layer_norm(key_value)
            target_tensor = input_tensor

        if key_value is None:
            key_value = input_tensor
        attention_output = self._attention_layer(
            query=target_tensor, value=key_value, attention_mask=attention_mask)
        attention_output = self._attention_dropout(attention_output)
        if self._norm_first:
            attention_output = source_tensor + attention_output
        else:
            attention_output = self._attention_layer_norm(target_tensor +
                                                          attention_output)
        if self._norm_first:
            source_attention_output = attention_output
            attention_output = self._output_layer_norm(attention_output)
        inner_output = self._intermediate_dense(attention_output)
        inner_output = self._intermediate_activation_layer(inner_output)
        inner_output = self._inner_dropout_layer(inner_output)
        layer_output = self._output_dense(inner_output)
        layer_output = self._output_dropout(layer_output)

        if self._norm_first:
            return source_attention_output + layer_output

        # During mixed precision training, layer norm output is always fp32 for now.
        # Casts fp32 for the subsequent add.
        layer_output = tf.cast(layer_output, tf.float32)
        return self._output_layer_norm(layer_output + attention_output)



from xdnlp.word_discover.ngram import Ngrams
from xdnlp.utils import logging_run_time, read_lines
from xdnlp.utils import default_logger as logging
from collections import Counter
from tqdm import tqdm
import os
import time
import math

settings = {
    "ngram": "ngram.pic",
    "corpus_name": "corpus.txt",
    "char_vocab_name": "char_vocab.txt",
    "ngram_pmi": "pmi.txt",
    "target_file": "new_words.vocab"
}

ngram = Ngrams()

class SimpleTrie:
    """Trie tree structure: used to search the continuous fragments made up of ngrams
    """

    def __init__(self):
        self.root = {}
        self.end = -1

    def insert(self, word):
        """Insert a word into the trie
        """
        curNode = self.root
        for c in word:
            if c not in curNode:
                curNode[c] = {}
            curNode = curNode[c]
        curNode[self.end] = True

    def tokenize(self, sentence):
        result = []
        start, end = 0, 1
        for i, c1 in enumerate(sentence):
            curNode = self.root
            if i == end:
                word = sentence[start: end]
                result.append(word)
                start, end = i, i + 1
            for j, c2 in enumerate(sentence[i:]):
                if c2 in curNode:
                    curNode = curNode[c2]
                    if self.end in curNode:
                        if i + j + 1 > end:
                            end = i + j + 1
                else:
                    break
        result.append(sentence[start: end])
        return result

    def search(self, word):
        curNode = self.root
        for c in word:
            if c not in curNode:
                return False
            curNode = curNode[c]
        # not end
        if self.end not in curNode:
            return False
        return True

    def max_match_cut(self, sentence, reverse=True):
        result = []
        i = len(sentence)
        if reverse:
            while i > 0:
                for j in range(0, i):
                    if self.search(sentence[j: i]) or i - j == 1:
                        result.append(sentence[j: i])
                        i = j
                        break
            return result[::-1]
        while i < len(sentence):
            max_index = i
            curNode = self.root
            for j in range(i, len(sentence)):
                if sentence[j] in curNode:
                    curNode = curNode[sentence[j]]
                    if self.end in curNode:
                        max_index = j
                else:
                    break
            result.append(sentence[i: max_index + 1])
            i = max_index + 1
        return result


class WordDiscover(object):

    def __init__(self, task_name="word-discover", min_pmi: list = [0, 2, 4, 6], min_count=32, pre_func=None):
        self.__run_at = str(int(time.time()))
        self.__save_path = f"./{task_name}-results/{self.__run_at}"
        self.__min_pmi = min_pmi
        self.__min_count = min_count
        self.__pre_func = pre_func
        self.__n = len(min_pmi)
        self.__candidate_trie = SimpleTrie()
        self.valid_ngrams = None
        self.candidates = None

    @logging_run_time
    def word_discover(self, filenames: list, remove_single_char=True, save_ngram=False):
        os.makedirs(self.__save_path)
        ngram.count_ngram(filenames, self.__n, self.__pre_func)
        if save_ngram:
            ngram.to_pickle(os.path.join(self.__save_path, settings["ngram"]))
        self.__filter_ngrams(ngram.ngram_list, ngram.total)
        self.__expansion_filter(filenames, remove_single_char=remove_single_char)
        self.__save()

    @logging_run_time
    def __filter_ngrams(self, ngrams, total):
        """filter the low pmi parts.
        Generate the candidate trie.
        """
        logging.info("Filter the ngrams by pmi, and create the candidate trie... ")
        output_ngrams = Counter()
        with open(os.path.join(self.__save_path, settings["ngram_pmi"]), 'w', encoding='utf-8') as pmi_file:
            for i in range(self.__n - 1, 0, -1):
                for w, v in ngrams[i].items():
                    pmi = min(
                        [total * v / (ngrams[j].get(w[:j + 1], v) * ngrams[i - j - 1].get(w[j + 1:], v))
                         for j in range(i)])
                    if math.log(pmi) >= self.__min_pmi[i]:
                        output_ngrams[w] = v
                        self.__candidate_trie.insert(w)
                    pmi_file.write(f"{w}\t{pmi}\n")
        self.valid_ngrams = output_ngrams

    def __expansion_filter(self, filenames, remove_single_char=True):
        """Extend the vocabulary using the longest match principle,and twice filter.
        """

        n_max = self.__n
        valid_ngrams = self.valid_ngrams

        def twice_filter(candidates):
            """filter the candidates keep the high mi ngrams
            """
            result = Counter()
            for i, j in candidates.items():
                if len(i) < 3:
                    result[i] = j
                elif len(i) <= n_max and i in valid_ngrams:
                    result[i] = j
                elif len(i) > n_max:
                    flag = True
                    for k in range(len(i) + 1 - n_max):
                        if i[k: k + n_max] not in valid_ngrams:
                            flag = False
                    if flag:
                        result[i] = j
            return result

        candidates = Counter()
        logging.info("Start generate the candidates...")
        for filename in filenames:
            for line in tqdm(read_lines(filename, self.__pre_func)):
                # line = "".join(line.strip().split())
                for w in self.__candidate_trie.tokenize(line.strip()):
                    # remove single char
                    if remove_single_char and len(w) < 2:
                        continue
                    candidates[w] += 1
        # filter by word frequency
        candidates = {i: j for i, j in candidates.items() if j >= self.__min_count}
        logging.info("Complete candidates by frequency.")

        # filter by high mutual information ngrams
        candidates = twice_filter(candidates)
        logging.info("Complete filter by mutual information of ngrams")
        logging.info("Generate word candidates successful!")
        self.candidates = candidates
        del candidates

    def __save(self):
        with open(os.path.join(self.__save_path, settings["target_file"]), 'w', encoding='utf-8') as f:
            for i, j in sorted(self.candidates.items(), key=lambda x: -x[1]):
                s = '%s %s\n' % (i, j)
                f.write(s)
        logging.info(f"Save vocab successfully: {os.path.join(self.__save_path, settings['target_file'])}")


# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Keras-based one-hot embedding layer."""
# pylint: disable=g-classes-have-attributes

import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="xdnlp")
class OnDeviceEmbedding(tf.keras.layers.Layer):
  """Performs an embedding lookup suitable for accelerator devices.

  This layer uses either tf.gather or tf.one_hot to translate integer indices to
  float embeddings.

  Args:
    vocab_size: Number of elements in the vocabulary.
    embedding_width: Output size of the embedding layer.
    initializer: The initializer to use for the embedding weights. Defaults to
      "glorot_uniform".
    use_one_hot: Whether to use tf.one_hot over tf.gather for the embedding
      lookup. Defaults to False (that is, using tf.gather). Setting this option
      to True may improve performance, especially on small vocabulary sizes, but
      will generally require more memory.
    scale_factor: Whether to scale the output embeddings. Defaults to None (that
      is, not to scale). Setting this option to a float will let values in
      output embeddings multiplied by scale_factor.
  """

  def __init__(self,
               vocab_size,
               embedding_width,
               initializer="glorot_uniform",
               use_one_hot=False,
               scale_factor=None,
               **kwargs):

    super(OnDeviceEmbedding, self).__init__(**kwargs)
    self._vocab_size = vocab_size
    self._embedding_width = embedding_width
    self._initializer = initializer
    self._use_one_hot = use_one_hot
    self._scale_factor = scale_factor

  def get_config(self):
    config = {
        "vocab_size": self._vocab_size,
        "embedding_width": self._embedding_width,
        "initializer": self._initializer,
        "use_one_hot": self._use_one_hot,
        "scale_factor": self._scale_factor,
    }
    base_config = super(OnDeviceEmbedding, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def build(self, input_shape):
    self.embeddings = self.add_weight(
        "embeddings",
        shape=[self._vocab_size, self._embedding_width],
        initializer=self._initializer,
        dtype=tf.float32)

    super(OnDeviceEmbedding, self).build(input_shape)

  def call(self, inputs):
    flat_inputs = tf.reshape(inputs, [-1])
    if self._use_one_hot:
      dtype = self._compute_dtype
      if not tf.dtypes.as_dtype(dtype).is_floating:
        # TensorFlow 1 compatibility. In TF1, self._compute_dtype is int32
        # instead of a floating-point dtype, as the dtype is inferred from the
        # dtype of the inputs
        dtype = tf.float32
      one_hot_data = tf.one_hot(
          flat_inputs, depth=self._vocab_size, dtype=dtype)
      embeddings = tf.matmul(one_hot_data, self.embeddings)
    else:
      embeddings = tf.gather(self.embeddings, flat_inputs)
    embeddings = tf.reshape(
        embeddings,
        # Work around b/142213824: prefer concat to shape over a Python list.
        tf.concat([tf.shape(inputs), [self._embedding_width]], axis=0))
    embeddings.set_shape(inputs.shape.as_list() + [self._embedding_width])
    if self._scale_factor:
      embeddings *= self._scale_factor
    return embeddings

  @property
  def vocab_size(self):
    return self._vocab_size

  @property
  def embedding_width(self):
    return self._embedding_width



import tensorflow as tf
import os
from xdnlp.utils import default_logger as logging


def load_data_from_directory(_path: str, batch_size, validation_split=0.1, seed=123, label_mode='categorical',
                             train=True):
    """train_dir: the train data dir
      test_dir: the test data dir
      Just set the directory:
      ```
      main_directory/
      ...class_a/
      ......a_text_1.txt
      ......a_text_2.txt
      ...class_b/
      ......b_text_1.txt
      ......b_text_2.txt
      ```
      """
    train_ds = None
    val_ds = None
    class_names = None
    if train:
        train_ds = tf.keras.preprocessing.text_dataset_from_directory(
            os.path.join(_path, 'train'), batch_size=batch_size, validation_split=validation_split,
            subset='training', seed=seed, label_mode=label_mode)
        class_names = train_ds.class_names
        train_ds = train_ds.cache().prefetch(tf.data.AUTOTUNE)

        val_ds = tf.keras.preprocessing.text_dataset_from_directory(
            os.path.join(_path, 'train'), batch_size=batch_size, validation_split=validation_split,
            subset='validation', seed=seed, label_mode=label_mode)
        val_ds = val_ds.cache().prefetch(tf.data.AUTOTUNE)
    test_ds = tf.keras.preprocessing.text_dataset_from_directory(
        os.path.join(_path, 'test'), batch_size=batch_size, label_mode=label_mode)
    if class_names is None:
        class_names = test_ds.class_names
    test_ds = test_ds.cache().prefetch(tf.data.AUTOTUNE)
    logging.info(f"Load data from directory successfully, class_names: {class_names}")
    return train_ds, val_ds, test_ds, class_names


def get_vectorize_layer(max_features, max_len, train_ds: tf.data.Dataset = None,
                        vocabulary=None, output_mode='int', split='whitespace') -> tf.keras.layers.TextVectorization:
    vectorize_layer = tf.keras.layers.TextVectorization(
        max_tokens=max_features,
        split=split,
        output_mode=output_mode,
        output_sequence_length=max_len,
        pad_to_max_tokens=True)
    if train_ds is not None:
        text_ds = train_ds.map(lambda x, y: x)
        vectorize_layer.adapt(text_ds)
    else:
        assert (vocabulary is not None, "if train_ds is None, vocabulary can not be None")
        vectorize_layer.adapt(tf.data.Dataset.from_tensor_slices(["just for init weights"]))
        vectorize_layer.set_vocabulary(vocabulary)
    logging.info(f"Generate vectorize layer successfully, and adapt: {vectorize_layer.is_adapted}")
    return vectorize_layer


def get_bert_tokenizer(vocab):
    lookup_table = tf.lookup.StaticVocabularyTable(
        tf.lookup.KeyValueTensorInitializer(
            keys=vocab,
            key_dtype=tf.string,
            values=tf.range(tf.size(vocab, out_type=tf.int64), dtype=tf.int64),
            value_dtype=tf.int64),
        num_oov_buckets=1,
        lookup_key_dtype=tf.string
    )


def train_format_data(filename):
    pass



from xdnlp.utils import md5, package_path, read_lines
import os


class Normalize(object):

    def __init__(self):
        self.__character_map = {}
        self.__chinese_map = {}
        self.__pinyin_map = {}
        self.__load_data()

    def normalize(self, query: str) -> str:
        result = ""
        for c in query:
            s = md5(c)
            c = self.__character_map.get(s, c)
            c = self.__chinese_map.get(s, c)
            result += c
        return result

    def pinyin(self, query: str, sep: str = " ") -> str:
        result = []
        for c in query:
            result.append(self.__pinyin_map.get(c, c))
        result = sep.join(result)
        return result

    def pinyin_with_space(self, query: str) -> str:
        result = []
        for c in query:
            result.append(self.__pinyin_map.get(c, c))
        return " ".join(result)

    def __load_data(self):
        for line in read_lines(os.path.join(package_path, "data/letter_mapping.txt")):
            temp = line.strip().split("\t")
            if len(temp) == 3:
                if temp[-1] == '0':
                    self.__character_map[temp[0]] = temp[1]
                elif temp[-1] == '1':
                    self.__chinese_map[temp[0]] = temp[1]

        for line in read_lines(os.path.join(package_path, "data/pinyin.txt")):
            temp = line.strip().split(",")
            if len(temp) >= 2:
                self.__pinyin_map[temp[0]] = temp[1]


if __name__ == '__main__':
    norm = Normalize()
    print(norm.normalize("""Nhận Cày Thuê Liên Quân Mobile Uy Tín - Giá Rẻ - Nhanh Chóng
Sdt/Zalo: 0363 099 883 liên hệ để biết thông tin chi tiết."""))



from nltk.util import ngrams
from collections import Counter
from xdnlp.utils import read_lines
import math
import joblib
import tqdm

class Gibberish(object):

    def __init__(self):
        self.counts = Counter()
        self.total = 0
        self.threshold = 0

    def adapt(self, filename: str, total=None):
        for line in tqdm.tqdm(read_lines(filename), total=total):
            for a, b in ngrams(line, 2):
                self.counts[a + b] += 1
                self.total += 1

    def adapt_threshold(self, goodfile: str, badfile: str):
        good = [self.average_prob(line) for line in read_lines(goodfile)]
        bad = [self.average_prob(line) for line in read_lines(badfile)]
        assert min(good) > max(bad)
        self.threshold = (min(good) + max(bad)) / 2

    def average_prob(self, query):
        assert self.total > 0, "Gibberish: should adapt first"
        pt = 0
        t = 0
        for a, b in ngrams(query, 2):
            p = self.counts.get(a + b, 1) / self.total
            pt += math.log(p)
            t += 1
        return pt / t

    def load(self, filename: str):
        t = joblib.load(filename)
        self.counts = t.get("counts", self.counts)
        self.total = t.get("total", self.total)
        self.threshold = t.get("threshold", self.threshold)

    def predict(self, query):
        assert self.total > 0, "Gibberish: should adapt first"
        assert self.threshold != 0, "Gibberish: should set threshold first"
        return self.average_prob(query) > self.threshold

    def save(self, filename):
        assert self.total > 0, "Gibberish: should adapt first"
        assert self.threshold != 0, "Gibberish: should set threshold first"
        joblib.dump({"counts": self.counts, "total": self.total, "threshold": self.threshold}, filename)


if __name__ == '__main__':
    gib = Gibberish()
    # gib.adapt("/home/geb/PycharmProjects/xdnlp/local/gibberish_data/data.txt")
    # joblib.dump({"counts": gib.counts, "total": gib.total}, "gib.model")
    gib.load("gib.model")
    e = gib.average_prob("ahsfuia")
    print(e)
    e = gib.average_prob("good")
    print(e)
    e = gib.average_prob("我是你爹")
    print(e)
    e = gib.average_prob("竺諸聞瀧韋")
    print(e)

    e = gib.average_prob("一起喵喵猫")
    print(e)

    e = gib.average_prob("熙熙菋菂夏天")
    print(e)

    e = gib.average_prob("碎花今如梦")
    print(e)

    e = gib.average_prob("若懷ほ稽")
    print(e)



from xdnlp.bert.layers.position_embedding import PositionEmbedding
from xdnlp.bert.layers.self_attention_mask import SelfAttentionMask
from xdnlp.bert.layers.on_device_embedding import OnDeviceEmbedding
from xdnlp.bert.layers.transformer_encoder_block import TransformerEncoderBlock



import collections
import tensorflow as tf
from xdnlp.bert.tf_utils import gelu
from xdnlp.bert.layers import PositionEmbedding, SelfAttentionMask, TransformerEncoderBlock, OnDeviceEmbedding


# @tf.keras.utils.register_keras_serializable(package='xdnlp')
class AlbertEncoder(tf.keras.Model):
    """ALBERT (https://arxiv.org/abs/1810.04805) text encoder network.

    This network implements the encoder described in the paper "ALBERT: A Lite
    BERT for Self-supervised Learning of Language Representations"
    (https://arxiv.org/abs/1909.11942).

    Compared with BERT (https://arxiv.org/abs/1810.04805), ALBERT refactorizes
    embedding parameters into two smaller matrices and shares parameters
    across layers.

    The default values for this object are taken from the ALBERT-Base
    implementation described in the paper.

    *Note* that the network is constructed by Keras Functional API.

    Args:
      vocab_size: The size of the token vocabulary.
      embedding_width: The width of the word embeddings. If the embedding width is
        not equal to hidden size, embedding parameters will be factorized into two
        matrices in the shape of `(vocab_size, embedding_width)` and
        `(embedding_width, hidden_size)`, where `embedding_width` is usually much
        smaller than `hidden_size`.
      hidden_size: The size of the transformer hidden layers.
      num_layers: The number of transformer layers.
      num_attention_heads: The number of attention heads for each transformer. The
        hidden size must be divisible by the number of attention heads.
      max_sequence_length: The maximum sequence length that this encoder can
        consume. If None, max_sequence_length uses the value from sequence length.
        This determines the variable shape for positional embeddings.
      type_vocab_size: The number of types that the 'type_ids' input can take.
      intermediate_size: The intermediate size for the transformer layers.
      activation: The activation to use for the transformer layers.
      dropout_rate: The dropout rate to use for the transformer layers.
      attention_dropout_rate: The dropout rate to use for the attention layers
        within the transformer layers.
      initializer: The initialzer to use for all weights in this encoder.
      dict_outputs: Whether to use a dictionary as the model outputs.
    """

    def __init__(self,
                 vocab_size,
                 embedding_width=128,
                 hidden_size=768,
                 num_layers=12,
                 num_attention_heads=12,
                 max_sequence_length=512,
                 type_vocab_size=2,
                 intermediate_size=3072,
                 activation=gelu,
                 dropout_rate=0.1,
                 attention_dropout_rate=0.1,
                 initializer=tf.keras.initializers.TruncatedNormal(stddev=0.02),
                 dict_outputs=True,
                 **kwargs):
        activation = tf.keras.activations.get(activation)
        initializer = tf.keras.initializers.get(initializer)

        self.hidden_size = hidden_size
        self.num_attention_heads = num_attention_heads

        word_ids = tf.keras.layers.Input(
            shape=(None,), dtype=tf.int32, name='input_word_ids')
        mask = tf.keras.layers.Input(
            shape=(None,), dtype=tf.int32, name='input_mask')
        type_ids = tf.keras.layers.Input(
            shape=(None,), dtype=tf.int32, name='input_type_ids')

        if embedding_width is None:
            embedding_width = hidden_size
        embedding_layer = OnDeviceEmbedding(
            vocab_size=vocab_size,
            embedding_width=embedding_width,
            initializer=initializer,
            name='word_embeddings')
        word_embeddings = embedding_layer(word_ids)

        # Always uses dynamic slicing for simplicity.
        position_embedding_layer = PositionEmbedding(
            initializer=initializer,
            max_length=max_sequence_length,
            name='position_embedding')
        position_embeddings = position_embedding_layer(word_embeddings)

        type_embeddings = (
            OnDeviceEmbedding(vocab_size=type_vocab_size,
                              embedding_width=embedding_width,
                              initializer=initializer,
                              use_one_hot=True,
                              name='type_embeddings')(type_ids))

        embeddings = tf.keras.layers.Add()(
            [word_embeddings, position_embeddings, type_embeddings])
        embeddings = (
            tf.keras.layers.LayerNormalization(
                name='embeddings/layer_norm',
                axis=-1,
                epsilon=1e-12,
                dtype=tf.float32)(embeddings))
        embeddings = (tf.keras.layers.Dropout(rate=dropout_rate)(embeddings))
        # We project the 'embedding' output to 'hidden_size' if it is not already
        # 'hidden_size'.
        if embedding_width != hidden_size:
            embeddings = tf.keras.layers.experimental.EinsumDense(
                '...x,xy->...y',
                output_shape=hidden_size,
                bias_axes='y',
                kernel_initializer=initializer,
                name='embedding_projection')(
                embeddings)

        data = embeddings
        attention_mask = SelfAttentionMask()(data, mask)
        shared_layer = TransformerEncoderBlock(
            num_attention_heads=num_attention_heads,
            inner_dim=intermediate_size,
            inner_activation=activation,
            output_dropout=dropout_rate,
            attention_dropout=attention_dropout_rate,
            kernel_initializer=initializer,
            name='transformer')
        encoder_outputs = []
        for _ in range(num_layers):
            data = shared_layer([data, attention_mask])
            encoder_outputs.append(data)

        # Applying a tf.slice op (through subscript notation) to a Keras tensor
        # like this will create a SliceOpLambda layer. This is better than a Lambda
        # layer with Python code, because that is fundamentally less portable.
        first_token_tensor = data[:, 0, :]
        cls_output = tf.keras.layers.Dense(
            units=hidden_size,
            activation='tanh',
            kernel_initializer=initializer,
            name='pooler_transform')(
            first_token_tensor)
        if dict_outputs:
            outputs = dict(
                sequence_output=data,
                encoder_outputs=encoder_outputs,
                pooled_output=cls_output,
            )
        else:
            outputs = [data, cls_output]

        # b/164516224
        # Once we've created the network using the Functional API, we call
        # super().__init__ as though we were invoking the Functional API Model
        # constructor, resulting in this object having all the properties of a model
        # created using the Functional API. Once super().__init__ is called, we
        # can assign attributes to `self` - note that all `self` assignments are
        # below this line.
        super(AlbertEncoder, self).__init__(
            inputs={'input_word_ids': word_ids, 'input_mask': mask, 'input_type_ids': type_ids}, outputs=outputs,
            **kwargs)
        config_dict = {
            'vocab_size': vocab_size,
            'embedding_width': embedding_width,
            'hidden_size': hidden_size,
            'num_layers': num_layers,
            'num_attention_heads': num_attention_heads,
            'max_sequence_length': max_sequence_length,
            'type_vocab_size': type_vocab_size,
            'intermediate_size': intermediate_size,
            'activation': tf.keras.activations.serialize(activation),
            'dropout_rate': dropout_rate,
            'attention_dropout_rate': attention_dropout_rate,
            'initializer': tf.keras.initializers.serialize(initializer),
        }

        # We are storing the config dict as a namedtuple here to ensure checkpoint
        # compatibility with an earlier version of this model which did not track
        # the config dict attribute. TF does not track immutable attrs which
        # do not contain Trackables, so by creating a config namedtuple instead of
        # a dict we avoid tracking it.
        config_cls = collections.namedtuple('Config', config_dict.keys())
        self._config = config_cls(**config_dict)
        self._embedding_layer = embedding_layer
        self._position_embedding_layer = position_embedding_layer

    def get_embedding_table(self):
        return self._embedding_layer.embeddings

    def get_config(self):
        return dict(self._config._asdict())

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def load_encoder_weights_from_google_ckpt(self, ckpt_path):
        variables = tf.train.load_checkpoint(ckpt_path)

        self.get_layer("word_embeddings").set_weights([
            variables.get_tensor("bert/embeddings/word_embeddings")])
        self.get_layer("position_embedding").set_weights([
            variables.get_tensor("bert/embeddings/position_embeddings")])
        self.get_layer("type_embeddings").set_weights([
            variables.get_tensor("bert/embeddings/token_type_embeddings")])

        self.get_layer("embeddings/layer_norm").set_weights([
            variables.get_tensor("bert/embeddings/LayerNorm/gamma"),
            variables.get_tensor("bert/embeddings/LayerNorm/beta")])

        self.get_layer("embedding_projection").set_weights([
            variables.get_tensor("bert/encoder/embedding_hidden_mapping_in/kernel"),
            variables.get_tensor("bert/encoder/embedding_hidden_mapping_in/bias")])

        self.get_layer("transformer").set_weights([
            tf.reshape(
                variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/attention_1/self/query/kernel"),
                [self.hidden_size, self.num_attention_heads, -1]),
            tf.reshape(
                variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/attention_1/self/query/bias"),
                [self.num_attention_heads, -1]),
            tf.reshape(
                variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/attention_1/self/key/kernel"),
                [self.hidden_size, self.num_attention_heads, -1]),
            tf.reshape(
                variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/attention_1/self/key/bias"),
                [self.num_attention_heads, -1]),
            tf.reshape(
                variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/attention_1/self/value/kernel"),
                [self.hidden_size, self.num_attention_heads, -1]),
            tf.reshape(
                variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/attention_1/self/value/bias"),
                [self.num_attention_heads, -1]),
            tf.reshape(
                variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/attention_1/output/dense/kernel"),
                [self.num_attention_heads, -1, self.hidden_size]),
            variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/attention_1/output/dense/bias"),
            variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/LayerNorm/gamma"),
            variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/LayerNorm/beta"),
            variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/ffn_1/intermediate/dense/kernel"),
            variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/ffn_1/intermediate/dense/bias"),
            variables.get_tensor(
                "bert/encoder/transformer/group_0/inner_group_0/ffn_1/intermediate/output/dense/kernel"),
            variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/ffn_1/intermediate/output/dense/bias"),
            variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/LayerNorm_1/gamma"),
            variables.get_tensor("bert/encoder/transformer/group_0/inner_group_0/LayerNorm_1/beta"),
        ])

        self.get_layer("pooler_transform").set_weights([
            variables.get_tensor("bert/pooler/dense/kernel"),
            variables.get_tensor("bert/pooler/dense/bias")])


if __name__ == '__main__':
    model = AlbertEncoder(vocab_size=21128)

    model.load_encoder_weights_from_google_ckpt("../data/albert_base/model.ckpt-best")
    # model.vocab_file = tf.saved_model.Asset("../data/albert_base/vocab_chinese.txt")
    # model.save("../data/hub_models/albert_base_zh", include_optimizer=False, save_format="tf")




tags = "!--...--,!DOCTYPE,a,abbr,acronym,address,applet,area,article,aside,audio,b,base,basefont,bdi,bdo,big,blockquote,body,br,button,canvas,caption,center,cite,code,col,colgroup,command,datalist,dd,del,details,dfn,dialog,dir,div,dl,dt,em,embed,fieldset,figcaption,figure,font,footer,form,frame,frameset,h1,h2,h3,h4,h5,h6,head,header,hr,html,i,iframe,img,input,ins,kbd,keygen,label,legend,li,link,main,map,mark,menu,meta,meter,nav,noframes,noscript,object,ol,optgroup,option,output,p,param,pre,progress,q,rp,rt,ruby,s,samp,script,section,select,small,source,span,strike,strong,style,sub,summary,sup,table,tbody,td,textarea,tfoot,th,thead,time,template,title,tr,track,tt,u,ul,var,video,wbr".split(
    ',')

tag2code = {}
for i, x in enumerate(tags):
    hex_str = hex(i).replace('0x', '')
    if len(hex_str) < 2:
        hex_str = '0' + hex_str

    tag2code[x] = hex_str


print(tag2code)


import pickle
import tqdm
from nltk.util import ngrams
from collections import Counter
from xdnlp.utils import read_lines, logging_run_time


class Ngrams(object):

    def __init__(self):
        self.ngram_list = []
        self.total = 0

    @logging_run_time
    def count_ngram(self, filenames: list, n: int, pre_func=None):
        ngram_list = [Counter() for _ in range(n)]
        total = 0
        for filename in filenames:
            for line in tqdm.tqdm(read_lines(filename, pre_func)):
                end_part = list(" " + line)
                if len(line) >= n:
                    sub_part_list = ngrams(line, n)
                    for sub_part in sub_part_list:
                        total += n
                        for i in range(n):
                            ngram_list[i]["".join(sub_part[:i + 1])] += 1
                    end_part = sub_part
                for i in range(1, l:=len(end_part)):
                    for j in range(i + 1, l + 1):
                        total += 1
                        ngram_list[j - i - 1]["".join(end_part[i:j])] += 1
        self.ngram_list = ngram_list
        self.total = total
        return ngram_list, total

    def to_pickle(self, filename: str):
        with open(filename, 'wb') as f:
            pickle.dump({
                'ngram_list': self.ngram_list,
                'total': self.total
            }, f)


if __name__ == '__main__':
    ng = Ngrams()

    # ngram_list, total = ng.count_ngram("/home/geb/PycharmProjects/word-discovery/data/ro_chat_data.txt", 4)
    # print(ngram_list)
    # ng.to_pickle("ng.pic")

    # from pybolt import bolt_nlp
    #
    # wd = bolt_nlp.WordDiscover()
    #
    # wd.word_discover(["/home/geb/PycharmProjects/word-discovery/data/ro_chat_data.txt"])



import os
import time
import joblib
import tensorflow as tf
import tensorflow_hub as hub
import tensorflow_text as text
from xdnlp.bert import optimization


class BertClassify(tf.keras.Model):
    """A text Classify use Bert or Albert pre training
    handle_encoder: the encoder model dir, saved_model format,
                    can use the model in tensorflow hub or local
    handle_preprocess: the preprocess model dir,saved_model format,
                    can use the model in tensorflow hub or local
    """

    def __init__(self,
                 handle_encoder,
                 handle_preprocess,
                 classifer_activation=None,
                 categories=2):

        self.handle_preprocess = handle_preprocess
        self.handle_encoder = handle_encoder

        text_input = tf.keras.layers.Input(shape=(), dtype=tf.string, name='text')
        preprocessing_layer = hub.KerasLayer(handle_preprocess, name='preprocessing')
        encoder_inputs = preprocessing_layer(text_input)
        encoder = hub.KerasLayer(handle_encoder, trainable=True, name='BERT_encoder')
        outputs = encoder(encoder_inputs)
        net = outputs['pooled_output']
        net = tf.keras.layers.Dropout(0.1)(net)
        net = tf.keras.layers.Dense(categories, activation=classifer_activation, name='classifier')(net)
        super(BertClassify, self).__init__(inputs=text_input, outputs=net)

    def preview_train_data(self, train_ds):
        for text_batch, label_batch in train_ds.take(1):
            for i in range(3):
                print(f'Review: {text_batch.numpy()[i]}')
                label = label_batch.numpy()[i]
                print(f'Label : {label} ({self.class_names[label]})')

    def preview_preprocess_encoder(self, txt: str):

        preprocess_model = hub.KerasLayer(self.handle_preprocess)
        encoder_model = hub.KerasLayer(self.handle_encoder)

        text_test = [txt]
        text_preprocessed = preprocess_model(text_test)
        print(f'Text       : {txt}')
        print(f'Keys       : {list(text_preprocessed.keys())}')
        print(f'Shape      : {text_preprocessed["input_word_ids"].shape}')
        print(f'Word Ids   : {text_preprocessed["input_word_ids"][0, :12]}')
        print(f'Input Mask : {text_preprocessed["input_mask"][0, :12]}')
        print(f'Type Ids   : {text_preprocessed["input_type_ids"][0, :12]}')
        bert_results = encoder_model(text_preprocessed)
        print(f'Loaded BERT: {self.handle_encoder}')
        print(f'Pooled Outputs Shape:{bert_results["pooled_output"].shape}')
        print(f'Pooled Outputs Values:{bert_results["pooled_output"][0, :12]}')
        print(f'Sequence Outputs Shape:{bert_results["sequence_output"].shape}')
        print(f'Sequence Outputs Values:{bert_results["sequence_output"][0, :12]}')

    def preview_classify(self, model):
        text_test = ['this is such an amazing movie!']
        bert_raw_result = model(tf.constant(text_test))
        print(f'raw result: {tf.sigmoid(bert_raw_result)}')

    def load_data(self,
                  train_dir,
                  test_dir,
                  batch_size=32,
                  validation_split=0.2,
                  seed=42,
                  buffer_size=tf.data.AUTOTUNE):
        """train_dir: the train data dir
        test_dir: the test data dir
        Just set the directory:
        ```
        main_directory/
        ...class_a/
        ......a_text_1.txt
        ......a_text_2.txt
        ...class_b/
        ......b_text_1.txt
        ......b_text_2.txt
        ```
        """
        raw_train_ds = tf.keras.preprocessing.text_dataset_from_directory(
            train_dir,
            batch_size=batch_size,
            validation_split=validation_split,
            subset='training',
            seed=seed)
        self.class_names = raw_train_ds.class_names
        train_ds = raw_train_ds.cache().prefetch(buffer_size=buffer_size)

        val_ds = tf.keras.preprocessing.text_dataset_from_directory(
            train_dir,
            batch_size=batch_size,
            validation_split=validation_split,
            subset='validation',
            seed=seed)
        val_ds = val_ds.cache().prefetch(buffer_size=buffer_size)

        test_ds = tf.keras.preprocessing.text_dataset_from_directory(
            test_dir,
            batch_size=batch_size)
        test_ds = test_ds.cache().prefetch(buffer_size=buffer_size)

        return train_ds, val_ds, test_ds

    def train(self, train_ds, val_ds, epochs=5, init_lr=3e-5, callbacks=[], verbose=1):
        loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
        steps_per_epoch = tf.data.experimental.cardinality(train_ds).numpy()
        num_train_steps = steps_per_epoch * epochs
        num_warmup_steps = int(0.1 * num_train_steps)
        optimizer = optimization.create_optimizer(init_lr=init_lr,
                                                  num_train_steps=num_train_steps,
                                                  num_warmup_steps=num_warmup_steps,
                                                  optimizer_type='adamw')
        save_path = os.path.join("models", str(int(time.time())))
        checkpoint_path = os.path.join(save_path, "cp-{epoch:04d}.ckpt")
        os.makedirs(save_path)
        joblib.dump(self.class_names, os.path.join(save_path, "label_name.pic"))
        if len(callbacks) == 0:
            cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath=checkpoint_path,
                                                             save_weights_only=True,
                                                             verbose=verbose,
                                                             save_freq="epoch")
        callbacks = [cp_callback]

        self.compile(optimizer=optimizer,
                     loss=loss,
                     metrics=["accuracy"])

        self.fit(train_ds,
                 validation_data=val_ds,
                 epochs=epochs,
                 callbacks=callbacks,
                 verbose=verbose)


if __name__ == '__main__':
    # model = BertClassify("https://tfhub.dev/tensorflow/small_bert/bert_en_uncased_L-4_H-512_A-8/1",
    #                      "https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/3",
    #                      categories=2)

    model = BertClassify("/home/geb/PycharmProjects/xdnlp/xdnlp/data/hub_models/albert_base_zh",
                         "/home/geb/PycharmProjects/xdnlp/xdnlp/data/hub_models/bert_zh_preprocess_3",
                         categories=2)

    train_ds, val_ds, test_ds = model.load_data("/home/geb/jonas/train", "/home/geb/jonas/test", )
    model.preview_train_data(train_ds)
    # model.preview_classify()
    model.train(train_ds, val_ds)

    loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)

    epochs = 5
    steps_per_epoch = tf.data.experimental.cardinality(model.train_ds).numpy()
    num_train_steps = steps_per_epoch * epochs
    num_warmup_steps = int(0.1 * num_train_steps)

    init_lr = 3e-5
    optimizer = optimization.create_optimizer(init_lr=init_lr,
                                              num_train_steps=num_train_steps,
                                              num_warmup_steps=num_warmup_steps,
                                              optimizer_type='adamw')

    model.compile(optimizer=optimizer,
                  loss=loss,
                  metrics=["accuracy"])

    history = model.fit(model.train_ds,
                        validation_data=model.val_ds,
                        epochs=epochs)



from xdnlp.text.prefix_set import PrefixSet
from xdnlp.text.normalize import Normalize
from xdnlp.text.encoder import FeatureEncoder
from joblib import Parallel, delayed
import multiprocessing
import re

n_cpus = multiprocessing.cpu_count()
default_pattern = re.compile("[^\u4E00-\u9FEF\u3400-\u4DB5a-zA-Z0-9]+", re.U)


class Text(object):

    def __init__(self):
        self.ps = PrefixSet()
        self.norm = Normalize()
        self.feature_encoder = FeatureEncoder()
        self.add_keyword = self.ps.add_keyword
        self.add_keywords_from_list = self.ps.add_keywords_from_list
        self.add_keywords_replace_map_from_dict = self.ps.add_keywords_replace_map_from_dict
        self.get_keywords = self.ps.get_keywords
        self.get_replace_map = self.ps.get_replace_map
        self.replace_keywords = self.ps.replace_keywords
        self.remove_keyword = self.ps.remove_keyword
        self.remove_keywords_from_list = self.ps.remove_keywords_from_list
        self.extract_keywords = self.ps.extract_keywords
        self.extract_keywords_with_index = self.ps.extract_keywords_with_index
        self.normalize = self.norm.normalize
        self.pinyin = self.norm.pinyin
        self.encode = self.feature_encoder.encode

    def clean(self, sentence, patten: str = None, keep_space: bool = True, norm: bool = True, lower: bool = True,
              digital_norm: bool = False, max_repeat: int = 0) -> str:
        if patten is None:
            patten = default_pattern
        else:
            patten = re.compile(patten, re.U)
        if norm:
            sentence = self.normalize(sentence)
        if lower:
            sentence = sentence.lower()
        sentence = patten.sub(" ", sentence)
        if keep_space:
            sentence = re.sub(r" +", " ", sentence)
        else:
            sentence = re.sub(r" +", "", sentence)
        if digital_norm:
            sentence = re.sub(r'\d+', '𝝢', sentence)
        if max_repeat > 0:
            sentence = re.sub(r'(.)\1{' + str(max_repeat) + r',}', r'\1' * max_repeat, sentence)
        return sentence

    def batch_extract_keywords(self, sentences, n_jobs=n_cpus, batch_size=1000):
        out = Parallel(n_jobs=n_jobs, verbose=9, batch_size=batch_size)(delayed(self.extract_keywords)(s) for s in sentences)
        return out

    def batch_replace_keywords(self, sentences, n_jobs=n_cpus, batch_size=1000):
        out = Parallel(n_jobs=n_jobs, verbose=9, batch_size=batch_size)(delayed(self.replace_keywords)(s) for s in sentences)
        return out

    def batch_clean(self, sentences, n_jobs=n_cpus, batch_size=1000, patten: str = None, keep_space: bool = True, norm: bool = True,
                    lower: bool = True, digital_norm: bool = False, max_repeat: int = 0):
        out = Parallel(n_jobs=n_jobs, verbose=9, batch_size=batch_size)(
            delayed(self.clean)(s, patten, keep_space, norm, lower, digital_norm, max_repeat) for s in sentences)
        return out

    def batch_cut(self, sentences, tokenizer, n_jobs=n_cpus, verbose=10, batch_size=1000):
        assert hasattr(tokenizer, 'lcut')

        def cut(s):
            return tokenizer.lcut(s)

        out = Parallel(n_jobs=n_jobs, verbose=verbose, batch_size=batch_size)(delayed(cut)(s) for s in sentences)
        return out



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Keras-based positional embedding layer."""
# pylint: disable=g-classes-have-attributes
import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="xdnlp")
class PositionEmbedding(tf.keras.layers.Layer):
    """Creates a positional embedding.

    Example:
    ```python
    position_embedding = PositionEmbedding(max_length=100)
    inputs = tf.keras.Input((100, 32), dtype=tf.float32)
    outputs = position_embedding(inputs)
    ```


    Args:
      max_length: The maximum size of the dynamic sequence.
      initializer: The initializer to use for the embedding weights. Defaults to
        "glorot_uniform".
      seq_axis: The axis of the input tensor where we add the embeddings.

    Reference: This layer creates a positional embedding as described in
    [BERT: Pre-training of Deep Bidirectional Transformers for Language
    Understanding](https://arxiv.org/abs/1810.04805).
    """

    def __init__(self,
                 max_length,
                 initializer="glorot_uniform",
                 seq_axis=1,
                 **kwargs):
        super(PositionEmbedding, self).__init__(**kwargs)
        if max_length is None:
            raise ValueError(
                "`max_length` must be an Integer, not `None`."
            )
        self._max_length = max_length
        self._initializer = tf.keras.initializers.get(initializer)
        self._seq_axis = seq_axis

    def get_config(self):
        config = {
            "max_length": self._max_length,
            "initializer": tf.keras.initializers.serialize(self._initializer),
            "seq_axis": self._seq_axis,
        }
        base_config = super(PositionEmbedding, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        dimension_list = input_shape.as_list()
        width = dimension_list[-1]
        weight_sequence_length = self._max_length

        self._position_embeddings = self.add_weight(
            "embeddings",
            shape=[weight_sequence_length, width],
            initializer=self._initializer)

        super(PositionEmbedding, self).build(input_shape)

    def call(self, inputs):
        input_shape = tf.shape(inputs)
        actual_seq_len = input_shape[self._seq_axis]
        position_embeddings = self._position_embeddings[:actual_seq_len, :]
        new_shape = [1 for _ in inputs.get_shape().as_list()]
        new_shape[self._seq_axis] = actual_seq_len
        new_shape[-1] = position_embeddings.get_shape().as_list()[-1]
        position_embeddings = tf.reshape(position_embeddings, new_shape)
        return tf.broadcast_to(position_embeddings, input_shape)





class PrefixSet(object):

    def __init__(self):
        self._prefix_dic = {}
        self._replace_map = {}

    def get_keywords(self):
        return {w for w, f in self._prefix_dic.items() if f == 1}

    def get_replace_map(self):
        return {a: b for a, b in self._replace_map.items() if b is not None}

    def add_keywords_from_list(self, words: list):
        for word in words:
            self.add_keyword(word)

    def add_keyword(self, word):
        w = ""
        for ch in word:
            w += ch
            if w not in self._prefix_dic:
                self._prefix_dic[w] = 0
        self._prefix_dic[w] = 1

    def add_keywords_replace_map_from_dict(self, source_target_map: dict):
        for a, b in source_target_map.items():
            w = ""
            for ch in a:
                w += ch
                if w not in self._replace_map:
                    self._replace_map[w] = None
            self._replace_map[a] = b

    def remove_keywords_from_list(self, words: list):
        for word in words:
            self.remove_keyword(word)

    def remove_keyword(self, word: str):
        self._prefix_dic[word] = 0

    def extract_keywords(self, sentence: str, longest_only=False) -> list:
        """Extract keywords involved in sentences
        Args:
            sentence: str, Sentences to be extracted.
            longest_only: bool,Whether to match only the longest keyword,default False;
                        for a example sentence: `category`, and keywords is ['cat', 'category'],
                        if set False, return: ['cat', 'category'],
                        if set True, return the longest only: ['category']
        """
        N = len(sentence)
        keywords = []
        for i in range(N):
            flag = sentence[i]
            j = i
            word = None
            while j < N and (flag in self._prefix_dic):
                if self._prefix_dic[flag] == 1:
                    if not longest_only:
                        keywords.append(flag)
                    else:
                        word = flag
                j += 1
                flag = sentence[i: j + 1]
            if longest_only and word:
                keywords.append(word)
        return keywords

    def extract_keywords_with_index(self, sentence: str, longest_only=False):
        N = len(sentence)
        keywords = []
        for i in range(N):
            flag, index = sentence[i], [i, i + 1]
            j = i
            word = None
            while j < N and (flag in self._prefix_dic):
                if self._prefix_dic[flag] == 1:
                    if not longest_only:
                        keywords.append((flag, index))
                    else:
                        word, _index = flag, index
                j += 1
                flag, index = sentence[i: j + 1], [i, j + 1]
            if longest_only and word:
                keywords.append((word, _index))
        return keywords

    def replace_keywords(self, sentence: str) -> str:
        """Replace word use keywords map.
        Args:
            sentence: str, Sentences that need to replace keywords.
        Return:
            the new sentence after replace keywords.
        """
        N = len(sentence)
        new_sentence = ""
        i = 0
        while i < N:
            flag = sentence[i]
            j = i
            word = None
            while j < N and (flag in self._replace_map):
                if self._replace_map[flag]:
                    word = flag
                j += 1
                flag = sentence[i: j + 1]
            if word:
                new_sentence += self._replace_map[word]
                i = j
            else:
                new_sentence += sentence[i]
                i += 1
        return new_sentence


# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Keras layer that creates a self-attention mask."""

import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package='xdnlp')
class SelfAttentionMask(tf.keras.layers.Layer):
    """Create 3D attention mask from a 2D tensor mask.

      inputs[0]: from_tensor: 2D or 3D Tensor of shape
        [batch_size, from_seq_length, ...].
      inputs[1]: to_mask: int32 Tensor of shape [batch_size, to_seq_length].

      Returns:
        float Tensor of shape [batch_size, from_seq_length, to_seq_length].
    """

    def call(self, inputs, to_mask):
        from_shape = tf.shape(inputs)
        batch_size = from_shape[0]
        from_seq_length = from_shape[1]

        to_shape = tf.shape(to_mask)
        to_seq_length = to_shape[1]

        to_mask = tf.cast(
            tf.reshape(to_mask, [batch_size, 1, to_seq_length]),
            dtype=inputs.dtype)

        # We don't assume that `from_tensor` is a mask (although it could be). We
        # don't actually care if we attend *from* padding tokens (only *to* padding)
        # tokens so we create a tensor of all ones.
        #
        # `broadcast_ones` = [batch_size, from_seq_length, 1]
        broadcast_ones = tf.ones(
            shape=[batch_size, from_seq_length, 1], dtype=inputs.dtype)

        # Here we broadcast along two dimensions to create the mask.
        mask = broadcast_ones * to_mask

        return mask



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bert encoder network."""
# pylint: disable=g-classes-have-attributes

import collections
from absl import logging
import tensorflow as tf
from xdnlp.bert import layers


# @tf.keras.utils.register_keras_serializable(package='xdnlp')
class BertEncoder(tf.keras.Model):
    """Bi-directional Transformer-based encoder network.

    This network implements a bi-directional Transformer-based encoder as
    described in "BERT: Pre-training of Deep Bidirectional Transformers for
    Language Understanding" (https://arxiv.org/abs/1810.04805). It includes the
    embedding lookups and transformer layers, but not the masked language model
    or classification task networks.

    The default values for this object are taken from the BERT-Base implementation
    in "BERT: Pre-training of Deep Bidirectional Transformers for Language
    Understanding".

    *Note* that the network is constructed by
    [Keras Functional API](https://keras.io/guides/functional_api/).

    Args:
      vocab_size: The size of the token vocabulary.
      hidden_size: The size of the transformer hidden layers.
      num_layers: The number of transformer layers.
      num_attention_heads: The number of attention heads for each transformer. The
        hidden size must be divisible by the number of attention heads.
      max_sequence_length: The maximum sequence length that this encoder can
        consume. If None, max_sequence_length uses the value from sequence length.
        This determines the variable shape for positional embeddings.
      type_vocab_size: The number of types that the 'type_ids' input can take.
      inner_dim: The output dimension of the first Dense layer in a two-layer
          feedforward network for each transformer.
      inner_activation: The activation for the first Dense layer in a two-layer
          feedforward network for each transformer.
      output_dropout: Dropout probability for the post-attention and output
          dropout.
      attention_dropout: The dropout rate to use for the attention layers
        within the transformer layers.
      initializer: The initialzer to use for all weights in this encoder.
      output_range: The sequence output range, [0, output_range), by slicing the
        target sequence of the last transformer layer. `None` means the entire
        target sequence will attend to the source sequence, which yields the full
        output.
      embedding_width: The width of the word embeddings. If the embedding width is
        not equal to hidden size, embedding parameters will be factorized into two
        matrices in the shape of ['vocab_size', 'embedding_width'] and
        ['embedding_width', 'hidden_size'] ('embedding_width' is usually much
        smaller than 'hidden_size').
      embedding_layer: An optional Layer instance which will be called to
       generate embeddings for the input word IDs.
      norm_first: Whether to normalize inputs to attention and intermediate
        dense layers. If set False, output of attention and intermediate dense
        layers is normalized.
    """

    def __init__(
            self,
            vocab_size,
            hidden_size=768,
            num_layers=12,
            num_attention_heads=12,
            max_sequence_length=512,
            type_vocab_size=16,
            inner_dim=3072,
            inner_activation=lambda x: tf.keras.activations.gelu(x, approximate=True),
            output_dropout=0.1,
            attention_dropout=0.1,
            initializer=tf.keras.initializers.TruncatedNormal(stddev=0.02),
            output_range=None,
            embedding_width=None,
            embedding_layer=None,
            norm_first=False,
            **kwargs):
        activation = tf.keras.activations.get(inner_activation)
        initializer = tf.keras.initializers.get(initializer)

        word_ids = tf.keras.layers.Input(
            shape=(None,), dtype=tf.int32, name='input_word_ids')
        mask = tf.keras.layers.Input(
            shape=(None,), dtype=tf.int32, name='input_mask')
        type_ids = tf.keras.layers.Input(
            shape=(None,), dtype=tf.int32, name='input_type_ids')

        if embedding_width is None:
            embedding_width = hidden_size

        if embedding_layer is None:
            embedding_layer_inst = layers.OnDeviceEmbedding(
                vocab_size=vocab_size,
                embedding_width=embedding_width,
                initializer=initializer,
                name='word_embeddings')
        else:
            embedding_layer_inst = embedding_layer
        word_embeddings = embedding_layer_inst(word_ids)

        # Always uses dynamic slicing for simplicity.
        position_embedding_layer = layers.PositionEmbedding(
            initializer=initializer,
            max_length=max_sequence_length,
            name='position_embedding')
        position_embeddings = position_embedding_layer(word_embeddings)
        type_embedding_layer = layers.OnDeviceEmbedding(
            vocab_size=type_vocab_size,
            embedding_width=embedding_width,
            initializer=initializer,
            use_one_hot=True,
            name='type_embeddings')
        type_embeddings = type_embedding_layer(type_ids)

        embeddings = tf.keras.layers.Add()(
            [word_embeddings, position_embeddings, type_embeddings])

        embedding_norm_layer = tf.keras.layers.LayerNormalization(
            name='embeddings/layer_norm', axis=-1, epsilon=1e-12, dtype=tf.float32)

        embeddings = embedding_norm_layer(embeddings)
        embeddings = (tf.keras.layers.Dropout(rate=output_dropout)(embeddings))

        # We project the 'embedding' output to 'hidden_size' if it is not already
        # 'hidden_size'.
        if embedding_width != hidden_size:
            embedding_projection = tf.keras.layers.experimental.EinsumDense(
                '...x,xy->...y',
                output_shape=hidden_size,
                bias_axes='y',
                kernel_initializer=initializer,
                name='embedding_projection')
            embeddings = embedding_projection(embeddings)
        else:
            embedding_projection = None

        transformer_layers = []
        data = embeddings
        attention_mask = layers.SelfAttentionMask()(data, mask)
        encoder_outputs = []
        self.num_layers = num_layers
        for i in range(num_layers):
            if i == num_layers - 1 and output_range is not None:
                transformer_output_range = output_range
            else:
                transformer_output_range = None
            layer = layers.TransformerEncoderBlock(
                num_attention_heads=num_attention_heads,
                inner_dim=inner_dim,
                inner_activation=inner_activation,
                output_dropout=output_dropout,
                attention_dropout=attention_dropout,
                norm_first=norm_first,
                output_range=transformer_output_range,
                kernel_initializer=initializer,
                name='transformer/layer_%d' % i)
            transformer_layers.append(layer)
            data = layer([data, attention_mask])
            encoder_outputs.append(data)

        last_encoder_output = encoder_outputs[-1]
        # Applying a tf.slice op (through subscript notation) to a Keras tensor
        # like this will create a SliceOpLambda layer. This is better than a Lambda
        # layer with Python code, because that is fundamentally less portable.
        first_token_tensor = last_encoder_output[:, 0, :]
        pooler_layer = tf.keras.layers.Dense(
            units=hidden_size,
            activation='tanh',
            kernel_initializer=initializer,
            name='pooler_transform')
        cls_output = pooler_layer(first_token_tensor)

        outputs = dict(
            sequence_output=encoder_outputs[-1],
            pooled_output=cls_output,
            encoder_outputs=encoder_outputs,
        )

        # Once we've created the network using the Functional API, we call
        # super().__init__ as though we were invoking the Functional API Model
        # constructor, resulting in this object having all the properties of a model
        # created using the Functional API. Once super().__init__ is called, we
        # can assign attributes to `self` - note that all `self` assignments are
        # below this line.
        super(BertEncoder, self).__init__(
            inputs={'input_word_ids': word_ids, 'input_mask': mask, 'input_type_ids': type_ids}, outputs=outputs,
            **kwargs)

        config_dict = {
            'vocab_size': vocab_size,
            'hidden_size': hidden_size,
            'num_layers': num_layers,
            'num_attention_heads': num_attention_heads,
            'max_sequence_length': max_sequence_length,
            'type_vocab_size': type_vocab_size,
            'inner_dim': inner_dim,
            'inner_activation': tf.keras.activations.serialize(activation),
            'output_dropout': output_dropout,
            'attention_dropout': attention_dropout,
            'initializer': tf.keras.initializers.serialize(initializer),
            'output_range': output_range,
            'embedding_width': embedding_width,
            'embedding_layer': embedding_layer,
            'norm_first': norm_first,
        }

        # We are storing the config dict as a namedtuple here to ensure checkpoint
        # compatibility with an earlier version of this model which did not track
        # the config dict attribute. TF does not track immutable attrs which
        # do not contain Trackables, so by creating a config namedtuple instead of
        # a dict we avoid tracking it.
        config_cls = collections.namedtuple('Config', config_dict.keys())
        self._config = config_cls(**config_dict)
        self._pooler_layer = pooler_layer
        self._transformer_layers = transformer_layers
        self._embedding_norm_layer = embedding_norm_layer
        self._embedding_layer = embedding_layer_inst
        self._position_embedding_layer = position_embedding_layer
        self._type_embedding_layer = type_embedding_layer
        if embedding_projection is not None:
            self._embedding_projection = embedding_projection

    def get_embedding_table(self):
        return self._embedding_layer.embeddings

    def get_embedding_layer(self):
        return self._embedding_layer

    def get_config(self):
        return dict(self._config._asdict())

    @property
    def transformer_layers(self):
        """List of Transformer layers in the encoder."""
        return self._transformer_layers

    @property
    def pooler_layer(self):
        """The pooler dense layer after the transformer layers."""
        return self._pooler_layer

    @classmethod
    def from_config(cls, config, custom_objects=None):
        if 'embedding_layer' in config and config['embedding_layer'] is not None:
            warn_string = (
                'You are reloading a model that was saved with a '
                'potentially-shared embedding layer object. If you contine to '
                'train this model, the embedding layer will no longer be shared. '
                'To work around this, load the model outside of the Keras API.')
            print('WARNING: ' + warn_string)
            logging.warn(warn_string)

        return cls(**config)

    def load_encoder_weights_from_google_ckpt(self, ckpt_path):
        variables = tf.train.load_checkpoint(ckpt_path)

        # embedding weights
        self._encoder_layer.get_layer("word_embeddings").set_weights([
            variables.get_tensor("bert/embeddings/word_embeddings")])
        self._encoder_layer.get_layer("position_embeddings").set_weights([
            variables.get_tensor("bert/embeddings/position_embeddings")])
        self._encoder_layer.get_layer("type_embeddings").set_weights([
            variables.get_tensor("bert/embeddings/token_type_embeddings")])

        self._encoder_layer.get_layer("embeddings/layer_norm").set_weights([
            variables.get_tensor("bert/embeddings/LayerNorm/gamma"),
            variables.get_tensor("bert/embeddings/LayerNorm/beta")
        ])

        # multi attention weights
        for i in range(self.num_layers):
            self._encoder_layer.get_layer("transformer/layer_{}".format(i)).set_weights([
                tf.reshape(variables.get_tensor(
                    "bert/encoder/layer_{}/attention/self/query/kernel".format(i)),
                    [self.bert_config.hidden_size, self.bert_config.num_attention_heads, -1]),
                tf.reshape(
                    variables.get_tensor("bert/encoder/layer_{}/attention/self/query/bias".format(i)),
                    [self.bert_config.num_attention_heads, -1]),
                tf.reshape(variables.get_tensor(
                    "bert/encoder/layer_{}/attention/self/key/kernel".format(i)),
                    [self.bert_config.hidden_size, self.bert_config.num_attention_heads, -1]),
                tf.reshape(
                    variables.get_tensor("bert/encoder/layer_{}/attention/self/key/bias".format(i)),
                    [self.bert_config.num_attention_heads, -1]),
                tf.reshape(variables.get_tensor(
                    "bert/encoder/layer_{}/attention/self/value/kernel".format(i)),
                    [self.bert_config.hidden_size, self.bert_config.num_attention_heads, -1]),
                tf.reshape(
                    variables.get_tensor("bert/encoder/layer_{}/attention/self/value/bias".format(i)),
                    [self.bert_config.num_attention_heads, -1]),
                tf.reshape(variables.get_tensor(
                    "bert/encoder/layer_{}/attention/output/dense/kernel".format(i)),
                    [self.bert_config.num_attention_heads, -1, self.bert_config.hidden_size]),
                variables.get_tensor("bert/encoder/layer_{}/attention/output/dense/bias".format(i)),
                variables.get_tensor("bert/encoder/layer_{}/attention/output/LayerNorm/gamma".format(i)),
                variables.get_tensor("bert/encoder/layer_{}/attention/output/LayerNorm/beta".format(i)),
                variables.get_tensor("bert/encoder/layer_{}/intermediate/dense/kernel".format(i)),
                variables.get_tensor("bert/encoder/layer_{}/intermediate/dense/bias".format(i)),
                variables.get_tensor("bert/encoder/layer_{}/output/dense/kernel".format(i)),
                variables.get_tensor("bert/encoder/layer_{}/output/dense/bias".format(i)),
                variables.get_tensor("bert/encoder/layer_{}/output/LayerNorm/gamma".format(i)),
                variables.get_tensor("bert/encoder/layer_{}/output/LayerNorm/beta".format(i)),
            ])

        self._encoder_layer.get_layer("pooler_transform").set_weights([
            variables.get_tensor("bert/pooler/dense/kernel"),
            variables.get_tensor("bert/pooler/dense/bias"),
        ])



import re
import os
from xdnlp.utils import package_path, read_lines
from xdnlp.text import Normalize, PrefixSet

norm = Normalize()


class FeatureEncoder(object):

    def __init__(self):
        self.feature_contact_pattern = r"([a-zA-Z\d,-,@,\.]{6,24})|[v微q扣]|(wx)|(weixin)"
        self.feature_unknown_pattern = r"[^\u0020-\u007e\u4e00-\u9fa5a-zA-Z\d]"
        self.feature_chinese_pattern = r"[\u0020-\u007e\u4e00-\u9fa5]"
        self.feature_en_num_pattern = r"[a-zA-Z0-9]"
        self.feature_not_chinese_pattern = r"[^\u0020-\u007e\u4e00-\u9fa5]"
        self.lfs = 1300
        self.prefix_set = PrefixSet()
        self.prefix_set.add_keywords_from_list(self._load_specify())
        self._chinese_frequency = self._load_chinese_frequency()

    def _load_specify(self):
        keywords = set()
        for line in read_lines(os.path.join(package_path, "data/ad.txt")):
            keywords.add(norm.pinyin(line.strip()))
        return list(keywords)

    def _load_chinese_frequency(self):
        chinese_frequency = {}
        for line in read_lines(os.path.join(package_path, "data/chinese_frequency.tsv")):
            t = line.split("\t")
            if len(t) > 1:
                chinese_frequency[t[0]] = int(t[1])
        return chinese_frequency

    def chinese_frequency(self, character: str) -> int:
        return self._chinese_frequency.get(character, 0)

    def chinese_frequency_list(self, sentence: str) -> list:
        return [self._chinese_frequency[c] for c in sentence if c in self._chinese_frequency]

    def feature_contact(self, text: str, pattern: str = None) -> int:
        if pattern is None:
            pattern = self.feature_contact_pattern
        if re.search(pattern, text):
            return 1
        return 0

    def feature_unknown(self, text: str, pattern: str = None) -> int:
        if pattern is None:
            pattern = self.feature_unknown_pattern
        if re.search(pattern, text):
            return 1
        else:
            return 0

    def feature_specify(self, text: str) -> int:
        if len(self.prefix_set.extract_keywords(text)) > 0:
            return 1
        return 0

    def feature_low_frequency(self, text: str) -> int:
        for c in text:
            if self._chinese_frequency.get(c, 20000) < 1300:
                return 1
        return 0

    def encode(self, text: str) -> dict:
        features = {'contact': self.feature_contact(text),
                    'unknown': self.feature_unknown(text),
                    'specify': self.feature_specify(text),
                    'length': len(text),
                    'low_frequency': 0,
                    'zh_scale': 0,
                    'en_num_scale': 0,
                    'zh_piece_scale': 0,
                    'not_zh_piece_scale': 0,
                    'pinyin': norm.pinyin(text, sep="")}
        total_count = 0
        zh_count = 0
        en_num_count = 0
        unknow_count = 0
        zh_pieces = re.split(self.feature_not_chinese_pattern, text)
        for character in text:
            if self._chinese_frequency.get(character, 10000) < self.lfs:
                features['low_frequency'] += 1
            if re.match(self.feature_chinese_pattern, character):
                zh_count += 1
                total_count += 1
            elif re.match(self.feature_en_num_pattern, character):
                en_num_count += 1/5
                total_count += 1/5
            elif re.match(self.feature_unknown_pattern, character):
                unknow_count += 1
                total_count += 1
        features["zh_scale"] = zh_count / total_count
        features["en_num_scale"] = en_num_count / total_count
        features["zh_piece_scale"] = len(zh_pieces) / (zh_count + 1)

        return features


if __name__ == '__main__':
    encoder = FeatureEncoder()

    print(encoder.feature_contact("v"))

    print(encoder.chinese_frequency_list("伽企鹅輑、フちqlフq、冲置二浙（格猫金、Zeny）套餐可选"))
    print(encoder.feature_unknown("伽企鹅輑、フちqlフq、冲置二浙（格猫金、Zeny）套餐可选"))
    print(encoder.feature_low_frequency("伽企鹅輑、フちqlフq、冲置二浙（格猫金、Zeny）套餐可选"))

    print(encoder.encode("wo操你妈、フちqlフq、"))


import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="xdnlp",
    version="0.0.12",
    author="geb",
    author_email="853934146@qq.com",
    description="An efficient NLP tool.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mikuh/xdnlp",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_data={'': ['data/*.txt', "data/*.tsv"]},
    install_requires=['numpy', 'pandas', 'nltk'],
    python_requires='>=3.6',
)


from re import I
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras import layers
import numpy as np
import os
import time
import joblib


class TextRNN(tf.keras.Model):

    def __init__(self,
                 vectorize_layer: layers.TextVectorization,
                 class_names: list,
                 model_dir: str = "models_TextRNN",
                 embedding_size: int = 128,
                 rnn_hidden_size: int = 256,
                 fc_hidden_size: int = 128,
                 num_layers: int = 2,
                 dropout: float = 0.2,
                 embedding_matrix_file: str = None,
                 embedding_matrix_trainable: bool = True,
                 save_model_config=True):
        embedding_matrix = np.load(embedding_matrix_file) if embedding_matrix_file is not None else None
        vocab_size = vectorize_layer.vocabulary_size()
        self.model_dir = os.path.join(model_dir, time.strftime(("%Y%m%d-%H%M%S"), time.localtime()))
        inputs = layers.Input(shape=(), dtype=tf.string, name='text')
        inputs_vec = vectorize_layer(inputs)
        if embedding_matrix is not None:
            x = layers.Embedding(vocab_size,
                                 embedding_size,
                                 embeddings_initializer=keras.initializers.Constant(self.embedding_matrix),
                                 trainable=embedding_matrix_trainable)(inputs_vec)
        else:
            x = layers.Embedding(vocab_size, embedding_size)(inputs_vec)

        for i in range(num_layers):
            x = layers.Bidirectional(layers.LSTM(rnn_hidden_size, return_sequences=True),
                                     name="BiLSTM_layer_%d" % (i))(x)
        x = RNN_Attention(name="attention_layer")(x)
        x = layers.Flatten()(x)
        x = layers.Dropout(dropout)(x)
        x = layers.Dense(fc_hidden_size, activation='relu', name="dense_layer_0")(x)
        outputs = layers.Dense(len(class_names), activation='softmax', name="dense_layer_1")(x)
        super(TextRNN, self).__init__(inputs=inputs, outputs=outputs, name="text_rnn")

        # save model config
        if save_model_config:
            if not os.path.exists(self.model_dir):
                os.makedirs(self.model_dir)
            joblib.dump(dict(class_names=list(class_names),
                             max_features=vocab_size,
                             max_len=inputs_vec.shape[1],
                             embedding_size=embedding_size,
                             rnn_hidden_size=rnn_hidden_size,
                             fc_hidden_size=fc_hidden_size,
                             num_layers=num_layers,
                             dropout=dropout,
                             vocabulary=list(vectorize_layer.get_vocabulary())),
                        os.path.join(self.model_dir, "model_config.pkl"))
        self.summary()

    @classmethod
    def get_model_config(cls, path):
        return joblib.load(os.path.join(path))

    def train(self, train_ds, val_ds, epochs=20, save_weights_only=True):
        checkpoint_path = os.path.join(self.model_dir, "checkpoint", "cp-{epoch:04d}.ckpt")
        cp_callback = keras.callbacks.ModelCheckpoint(filepath=checkpoint_path,
                                                      save_weights_only=save_weights_only,
                                                      verbose=1,
                                                      save_freq="epoch")
        log_dir = os.path.join(self.model_dir, "logs")
        tensorboard_callback = keras.callbacks.TensorBoard(log_dir=log_dir,
                                                           update_freq=1000,
                                                           write_graph=False)

        self.compile(optimizer='adam',
                     loss='categorical_crossentropy',
                     metrics=['accuracy'])

        self.fit(train_ds,
                 validation_data=val_ds,
                 epochs=epochs,
                 callbacks=[cp_callback, tensorboard_callback])

        self.save(os.path.join(self.model_dir, "my_model"))


class RNN_Attention(layers.Layer):
    def __init__(self, **kwargs):
        super(RNN_Attention, self).__init__(**kwargs)
        self.attention = layers.Attention()

    def build(self, input_shape):
        self.w = self.add_weight(shape=(1, input_shape[-1]),
                                 initializer="random_normal",
                                 trainable=True,
                                 name="attention_query")

    def call(self, inputs):
        outputs = self.attention([self.w, inputs])
        return outputs


import tensorflow as tf
import numpy as np
import os
import datetime
from tensorflow import keras
import joblib

class TextCNN(tf.keras.Model):

    def __init__(self, input_shape: tuple,
                 class_names: list,
                 vectorize_layer: keras.layers.TextVectorization,
                 model_dir: str = "models",
                 embedding_size: int = 128,
                 hidden_size: int = 256,
                 filter_sizes: list = [3, 4, 5],
                 num_filters: int = 256,
                 dropout: float = 0.2,
                 embedding_matrix_file: str = None,
                 embedding_matrix_trainable: bool = True,
                 is_train=True):
        embedding_matrix = np.load(embedding_matrix_file) if embedding_matrix_file is not None else None
        vocab_size = vectorize_layer.vocabulary_size()
        inputs = keras.Input(shape=(), dtype=tf.string, name='text')
        inputs_vec = vectorize_layer(inputs)
        if embedding_matrix is not None:
            x = keras.layers.Embedding(vocab_size,
                                       embedding_size,
                                       embeddings_initializer=tf.keras.initializers.Constant(self.embedding_matrix),
                                       trainable=embedding_matrix_trainable)(inputs_vec)
        else:
            x = keras.layers.Embedding(vocab_size, embedding_size)(inputs_vec)
        xs = [
            keras.layers.Conv1D(num_filters, filter_size, activation='relu', input_shape=input_shape, padding='same',
                                name=f"cnn_size_{filter_size}")(x)
            for filter_size in filter_sizes]
        xs = [keras.layers.MaxPool1D(pool_size=input_shape[0], name=f"max_pool_{i}")(x) for i, x in enumerate(xs)]
        x = keras.layers.Concatenate(axis=-1, name="concatenate")(xs)
        x = keras.layers.Flatten()(x)
        x = keras.layers.Dropout(dropout)(x)
        x = keras.layers.Dense(hidden_size, activation='relu', name="hidden_layer")(x)
        x = keras.layers.Dense(len(class_names), activation='softmax', name="output")(x)
        super(TextCNN, self).__init__(inputs=inputs, outputs=x, name="text_cnn")
        # save model config
        if is_train:
            self.model_dir = os.path.join(model_dir, datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
            if not os.path.exists(self.model_dir):
                os.makedirs(self.model_dir)
            joblib.dump(dict(class_names=list(class_names),
                             max_features=vocab_size,
                             max_len=input_shape[0],
                             embedding_size=embedding_size,
                             hidden_size=hidden_size,
                             filter_sizes=filter_sizes,
                             num_filters=num_filters,
                             dropout=dropout,
                             vocabulary=list(vectorize_layer.get_vocabulary())),
                        os.path.join(self.model_dir, "model_config.pkl"))
        self.summary()

    @classmethod
    def get_model_config(cls, path):
        return joblib.load(os.path.join(path, "model_config.pkl"))

    def train(self, train_ds, val_ds, epochs=20, save_weights_only=True):
        checkpoint_path = os.path.join(self.model_dir, "cp-{epoch:04d}.ckpt")
        cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath=checkpoint_path,
                                                         save_weights_only=save_weights_only,
                                                         verbose=1,
                                                         save_freq="epoch")
        log_dir = os.path.join(self.model_dir, "logs/")
        tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir)

        self.compile(optimizer='adam',
                     loss='categorical_crossentropy',
                     metrics=['accuracy'])

        self.fit(train_ds,
                 validation_data=val_ds,
                 epochs=epochs,
                 callbacks=[cp_callback, tensorboard_callback])

        self.save(os.path.join(self.model_dir, "my_model"))



from sklearn.feature_extraction.text import CountVectorizer
from sklearn.naive_bayes import BernoulliNB
import os
import joblib
import time
import numpy as np
import json
from xdnlp.utils import default_logger as logging
from xdnlp.utils import read_csv_row


class BernouliNB(object):
    def __init__(self, alpha=0.1, is_train=True, preprocess=None):
        self.vec_model = CountVectorizer(lowercase=False, token_pattern=r"(?u)[^ \n]+")
        self.classification_model = BernoulliNB(alpha=alpha)
        self.__save_path = "./models/" + str(int(time.time()))
        self.accuracy = 0.0
        if preprocess is None:
            self.preprocess = lambda x: x.split()
        else:
            self.preprocess = preprocess
        if not os.path.exists(self.__save_path) and is_train:
            os.makedirs(self.__save_path)

    def fit(self, inputs, targets):
        inputs = self.vec_model.fit_transform(inputs)
        self.classification_model.fit(inputs, targets)
        self.accuracy = self.__score(inputs, targets)
        logging.info(f"\nAccuracy: {self.accuracy}")

        self.__save()
        logging.info(f"Save model and parameters, at: {self.__save_path}")

    def __score(self, inputs, targets):
        return self.classification_model.score(inputs, targets)

    def __save(self):
        joblib.dump(self.vec_model, os.path.join(self.__save_path, "vectorizer.model"))
        joblib.dump(self.classification_model, os.path.join(self.__save_path, "clf_bernoulli_nb.model"))
        joblib.dump(self.preprocess, os.path.join(self.__save_path, "preprocess.model"))
        self.export_parameter()

    def load(self, model_path):
        self.vec_model = joblib.load(os.path.join(model_path, "vectorizer.model"))
        self.classification_model = joblib.load(os.path.join(model_path, "clf_bernoulli_nb.model"))

    def load_data(self, filename):
        targets = []
        inputs = []
        for row in read_csv_row(filename):
            if len(row) == 3:
                for _ in range(int(row[2])):
                    targets.append(row[0])
                    inputs.append(self.preprocess(row[1]))
        return inputs, targets

    def predict(self, inputs):
        count_vec = self.vec_model.transform(inputs)
        ret = list(self.classification_model.predict(count_vec))
        return ret

    def export_parameter(self):
        features = self.vec_model.get_feature_names()
        vocab = self.vec_model.vocabulary_
        classes_ = self.classification_model.classes_
        pos_log_prob = self.classification_model.feature_log_prob_
        neg_prob = np.log(1 - np.exp(self.classification_model.feature_log_prob_))
        del_prob = (pos_log_prob - neg_prob).T
        class_log_prob = self.classification_model.class_log_prior_

        json_data = {
            "Classes": classes_.tolist(),
            "Features": features,
            "Vocab": {k: int(v) for k, v in vocab.items()},
            "DelProb": del_prob.tolist(),
            "NegProb": neg_prob.tolist(),
            "ClassLogProb": class_log_prob.tolist()
        }

        with open(os.path.join(self.__save_path, "model.json"), 'w', encoding='utf-8') as f:
            json.dump(json_data, f)
